import requests, validators
from bs4 import BeautifulSoup


class Preloader:
    """
    Preloader class to fetch all URLs from a sitemap and its sitemap index"""

    def __init__(self, sitemap_url, depth=2, allow_redirects=False):
        """
        :param sitemap_url: URL of the sitemap
        :param depth: Depth of the sitemap index
        """
        self.sitemap_url = sitemap_url
        self.sitemap_urls = set()
        self.page_urls = set()
        self.failed_urls = {}
        self.finished_pages = []
        self.original_pages = set()
        self.allow_redirects = allow_redirects
        if depth < 1:
            raise ValueError("depth must be greater than 1")
        self.depth = depth
        self.extract_urls()

    def extract_urls(self):
        """
        Extracts all URLs from the sitemap and store them in a set
        """
        self.fetch_url(self.sitemap_url)
        print(f"Found {len(self.sitemap_urls)} sitemaps")
        print(f"Found {len(self.page_urls)} pages")
        self.original_pages = self.page_urls.copy()

    def fetch_pages(self, batch_size=None):
        """
        Fetches extracted urls
        """
        total_pages = len(self.page_urls)
        fetched_pages = 0

        print(
            f"Will process {batch_size if batch_size else total_pages} pages of {total_pages}"
        )

        page_urls = list(self.page_urls)
        pages_range = page_urls[:batch_size] if batch_size else page_urls
        for url in pages_range:
            fetched_pages += 1
            print(
                f"Fetching page: {fetched_pages}/{batch_size if batch_size else total_pages}"
            )
            self.fetch_url(url, self.depth)

    def fetch_url(self, sitemap_url, level=0):
        """
        Fetches all URLs from a sitemap or sitemap index
        :param sitemap_url: URL of the sitemap or sitemap index
        :param level: Level of the sitemap index
        """
        print(f"Fetching: {sitemap_url}")
        response = requests.get(sitemap_url, allow_redirects=self.allow_redirects)
        urls = []
        if response.status_code == 200:
            # Don't analyze it if It's the latest level
            if level < self.depth:
                soup = BeautifulSoup(response.text, "xml")  # Use lxml as the parser
                urls = [loc.text for loc in soup.find_all("loc")]

                if level < self.depth - 1:
                    for url in urls:
                        self.add_url(self.sitemap_urls, url)
                        self.fetch_url(url, 1)
                elif level == self.depth - 1:
                    for url in urls:
                        self.add_url(self.page_urls, url)
            else:
                self.finished_pages.append(sitemap_url)
                self.page_urls.remove(sitemap_url)

        else:
            print(f"Failed to fetch URL: {response.status_code}")
            if response.status_code in self.failed_urls:
                self.failed_urls[response.status_code].append(sitemap_url)
            else:
                self.failed_urls[response.status_code] = [sitemap_url]

            if level == self.depth:
                self.page_urls.remove(sitemap_url)
            else:
                self.sitemap_urls.remove(sitemap_url)

    def add_url(self, base_set, url):
        """
        Adds a URL to the provided set if it's valid only
        :param base_set: Set to add the URL to
        :param url: URL to add to the set
        """
        if validators.url(url):
            base_set.add(url)



from setuptools import setup, find_packages

# Package metadata
name = "sitemap-preloader"
version = "0.1.3"
description = "This package is used to read a sitemap and fetch all the urls in order to warm the cache of a website"
author = "Andres Gonzalez"
author_email = "code@andresgz.com"
url = "https://github.com/andresgz/sitemap-preloader"

classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]

install_requires = [
    "requests",
    "beautifulsoup4",
    "validators",
    "lxml",
]

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name=name,
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=author,
    author_email=author_email,
    url=url,
    py_modules=["preloader"],
    classifiers=classifiers,
    install_requires=install_requires,
)



import sys
import os

# Add the project's root directory to the Python path
project_root = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, project_root)

import pytest, requests
from preloader import Preloader
from unittest.mock import patch, Mock

# Define the path to the external mock XML file
MOCK_XML_FILE_MAIN = os.path.join(os.path.dirname(__file__), "mock_sitemap_main.xml")
MOCK_XML_FILE_MAIN_ERROR = os.path.join(
    os.path.dirname(__file__), "mock_sitemap_main_error.xml"
)
MOCK_XML_FILE_SUB1 = os.path.join(os.path.dirname(__file__), "mock_sitemap_sub1.xml")
MOCK_XML_FILE_SUB2 = os.path.join(os.path.dirname(__file__), "mock_sitemap_sub2.xml")
MOCK_XML_FILE_SUB_ERROR = os.path.join(
    os.path.dirname(__file__), "mock_sitemap_sub_error.xml"
)
MOCK_PAGE = os.path.join(os.path.dirname(__file__), "page.html")


@pytest.fixture
def mock_requests_get():
    """Mock requests.get to return our own XML file for each URL"""

    def custom_get(url, *args, **kwargs):
        response = requests.Response()
        if url == "https://example.com/sitemap.xml":
            response.status_code = 200
            with open(MOCK_XML_FILE_MAIN, "rb") as mock_file:
                response._content = mock_file.read()
        elif url == "https://example.com/sitemap_error.xml":
            response.status_code = 200
            with open(MOCK_XML_FILE_MAIN_ERROR, "rb") as mock_file:
                response._content = mock_file.read()
        elif url == "https://example.com/sitemap_sub1.xml":
            response.status_code = 200
            with open(MOCK_XML_FILE_SUB1, "rb") as mock_file:
                response._content = mock_file.read()
        elif url == "https://example.com/sitemap_sub2.xml":
            response.status_code = 200
            with open(MOCK_XML_FILE_SUB2, "rb") as mock_file:
                response._content = mock_file.read()
        elif url == "https://example.com/sitemap_sub_error.xml":
            response.status_code = 200
            with open(MOCK_XML_FILE_SUB_ERROR, "rb") as mock_file:
                response._content = mock_file.read()
        elif (
            url == "https://example.com/page1.html"
            or url == "https://example.com/page2.html"
            or url == "https://example.com/page3.html"
            or url == "https://example.com/page4.html"
            or url == "https://example.com/page5.html"
            or url == "https://example.com/page6.html"
        ):
            response.status_code = 200
            with open(MOCK_PAGE, "rb") as mock_file:
                response._content = mock_file.read()
        else:
            response.status_code = 404
        return response

    with patch("requests.get", side_effect=custom_get) as mock_get:
        yield mock_get


def test_mock_ok(mock_requests_get):
    """Test that the mock is returning the file content"""
    response = requests.get("https://example.com/sitemap.xml")
    assert response.status_code == 200
    with open(MOCK_XML_FILE_MAIN, "rb") as file:
        expected_content = file.read()
    assert response.content == expected_content


def test_mock_fail(mock_requests_get):
    """Test that the mock is returning 404 for other URLs"""
    response = requests.get("https://example.com/other.xml")
    assert response.status_code == 404


def test_value_error(mock_requests_get):
    with pytest.raises(ValueError) as exc_info:
        Preloader("https://example.com/sitemap.xml", depth=0)
    assert str(exc_info.value) == "depth must be greater than 1"


def test_preloader_fetch_level_1(mock_requests_get):
    """Test a 1 level sitemap"""
    preloader = Preloader("https://example.com/sitemap.xml", depth=1)
    assert len(preloader.page_urls) == 2
    assert len(preloader.sitemap_urls) == 0


def test_preloader_level_2(mock_requests_get):
    """Test a 2 levels sitemap"""
    preloader = Preloader("https://example.com/sitemap.xml", depth=2)
    assert preloader.sitemap_url == "https://example.com/sitemap.xml"
    assert len(preloader.page_urls) == 6
    assert len(preloader.sitemap_urls) == 2


def test_preloader_level_3(mock_requests_get):
    """Test a 3 levels sitemap"""
    preloader = Preloader("https://example.com/sitemap.xml", depth=3)
    assert preloader.sitemap_url == "https://example.com/sitemap.xml"
    assert len(preloader.page_urls) == 0
    assert len(preloader.sitemap_urls) == 8


def test_fetch_all_pages(mock_requests_get):
    """Test that the preloader can fetch all pages from a sitemap"""
    preloader = Preloader("https://example.com/sitemap.xml", depth=2)
    assert len(preloader.page_urls) == 6
    preloader.fetch_pages()
    assert len(preloader.finished_pages) == 6
    assert len(preloader.page_urls) == 0
    assert len(preloader.sitemap_urls) == 2


def test_fetch_pages_batch(mock_requests_get):
    """Test that the preloader can fetch pages in batches"""
    preloader = Preloader("https://example.com/sitemap.xml", depth=2)
    assert len(preloader.page_urls) == 6
    preloader.fetch_pages(3)
    assert len(preloader.finished_pages) == 3
    assert len(preloader.page_urls) == 3
    preloader.fetch_pages(3)
    assert len(preloader.finished_pages) == 6
    assert len(preloader.page_urls) == 0


def test_fetch_sitemap_with_error_page(mock_requests_get):
    """
    Test that the preloader can handle a sitemap with iteams returning an error page
    """
    preloader = Preloader("https://example.com/sitemap_error.xml", depth=2)
    assert len(preloader.sitemap_urls) == 1
    assert len(preloader.failed_urls) == 1
    assert len(preloader.page_urls) == 3
    preloader.fetch_pages()
    assert len(preloader.failed_urls[404]) == 3


if __name__ == "__main__":
    pytest.main()
